/***********************************************************************/
/*                                                                     */
/*                                                                     */
/* Program:                                                            */
/* --------                                                            */
/*                                                                     */
/* ssei.c,  Calculates e = exp(1), Integer method.                     */
/*                                                                     */
/*                                                                     */
/* Version:                                                            */
/* --------                                                            */
/*                                                                     */
/* First version: 1998-02-16                                           */
/* Latest update: 1998-06-22                                           */
/*                                                                     */
/*                                                                     */
/* Author of the program:                                              */
/* ----------------------                                              */
/*                                                                     */
/* Stefan Spaennare  Lund, Sweden                                      */
/*                                                                     */
/* Email: stefans@astro.lu.se                                          */
/*                                                                     */
/* If you have problems with the program please let me know !          */
/*                                                                     */
/*                                                                     */
/* Copy-right:                                                         */
/* -----------                                                         */
/*                                                                     */
/* This program is free to copy if this information follows the        */
/* program.                                                            */
/*                                                                     */
/*                                                                     */
/* IMPORTANT !                                                         */
/* -----------                                                         */
/*                                                                     */
/* Please don't use the result from this program for important         */
/* purposes without doing an independent check of the result !         */
/* See also the notice below !                                         */
/*                                                                     */
/* OBSERVE ! This program probably gives wrong result if n > 10^6 !    */
/*                                                                     */
/*                                                                     */
/* Function:                                                           */
/* ---------                                                           */
/*                                                                     */
/* This C program calculates e = exp(1) to many decimal places.        */
/*                                                                     */
/* The parameter ni <= 10^6 is the number of decimals.                 */
/*                                                                     */
/*                                                                     */
/* Compilation:                                                        */
/* ------------                                                        */
/*                                                                     */
/* The program works in UNIX and Linux and I recommend that you        */
/* optimize the program (i.e. use cc -O2 option or similar option)     */
/* when you compile (gcc if you use a GNU compiler).                   */
/*                                                                     */
/*> cc -O2 -o ssei ssei.c -lm                                          */
/*                                                                     */
/* For some UNIX dialects the time measurement doesn't work.           */
/* Then change "CLK_TCK" to something else (1.0e6 on HP and            */
/* some other computers). The time wraps around after 4295             */
/* seconds on some (HP) computers.                                     */ 
/*                                                                     */
/*                                                                     */
/* Output:                                                             */
/* -------                                                             */
/*                                                                     */
/* Output is to the file "ssei.dat". This file name can easily be      */
/* changed.                                                            */
/*                                                                     */
/*                                                                     */
/* Algorithm and accuracy:                                             */
/* -----------------------                                             */
/*                                                                     */
/* The algorithm is from a student exercise in Pascal programming      */
/* at Lund Institute of Technology in February 1998.                   */
/*                                                                     */
/* The result has been verified up to 10^6 decimals.                   */
/*                                                                     */
/* OBSERVE ! This program probably gives wrong result if n > 10^6 !    */
/*                                                                     */
/* The base for calculations is 10^4 (short arrays).                   */
/*                                                                     */
/* The program code is optimized for speed by the author.              */
/*                                                                     */
/*                                                                     */
/* Memory:                                                             */
/* -------                                                             */
/*                                                                     */
/* The program requires approximatley 1.0 Mbyte memory per million     */
/* digits of e.                                                        */
/*                                                                     */
/*                                                                     */
/* Timing and benchmarks:                                              */
/* ----------------------                                              */
/*                                                                     */
/* The time depends on the factorial that fits into the n digits.      */
/* Empirically this means that the computing time approximately        */
/* increases with a factor of four if the number of digits doubles.    */
/*                                                                     */
/*                                                                     */
/* The program is well adapted for benchmark testing and               */
/* has been tested on two different computers:                         */
/*                                                                     */
/* 1: A DELL DIMENSION XPS P90 computer with a 180 MHz Pentium MMX     */
/* overdrive processor, 96 Mbyte memory and 256 kbyte cache.           */
/* The operating system was Linux Red Hat 4.1 and the code was         */
/* fully optimized (gcc -O3).                                          */
/*                                                                     */
/* 2: A HP 735 computer (99 MHz) with 144 Mbyte memory and 256 kbyte   */
/* cache. The operating system was HP UNIX and the code was fully      */
/* optimized (cc +O4).                                                 */ 
/*                                                                     */
/* The CPU-times for calculations of e follows in the table below.     */
/* Observe that the times can vary a few percent depending load etc.   */
/*                                                                     */
/* ------------------------------------------------------------------  */
/*                P 180 MMX OD       HP 735                            */
/* Digits       (s)  (h  m  s)     (s)  (h  m  s)                      */
/* ------------------------------------------------------------------- */
/* 10^5         170  00 02 50      165  00 02 45                       */
/* 10^6       15223  04 13 43         Not calc.                        */
/* ------------------------------------------------------------------- */
/*                                                                     */
/***********************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Notice:                                                                   */
/* =======                                                                   */
/*                                                                           */
/* I make no warranties that this program is (1) free of error, (2) con-     */
/* sistent with any standard merchantability, or (3) meeting the require-    */
/* ments of a particular application.  This software shall not, partly or    */
/* as a whole, participate in a process, whose outcome can result in injury  */
/* to a person or loss of property. It is solely designed for analytical     */
/* work.  Permission to use, copy, and distribute is hereby granted without  */
/* fee, providing that the header above including this notice appears in     */
/* all copies.                                                               */
/*                                                                           */
/*                                                         Stefan Spaennare  */
/*                                                                           */
/*****************************************************************************/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

double timetic=(double)(CLOCKS_PER_SEC);

char fname[80]="ssei.dat";

void fppp(t,outfile)
int t; 
FILE *outfile;

{
  if (t > 999) {
     fprintf(outfile,"%4d",t);
  }
  else if (t > 99) {
     fprintf(outfile,"0%3d",t);
  } 
  else if (t > 9) {
     fprintf(outfile,"00%2d",t);
  }
  else {
     fprintf(outfile,"000%1d",t);
  } /* if */

} /* fppp */

int main(argc,argv)
int argc;
char *argv[];

{
   
   FILE *outfile;

   int i,j,n,l1,l2,rem,q,fac;
   int imin,il,ll,ll2,bas;

   double lt,l2f,sum,jd,lld;

   clock_t time1,timediff1;

   short *v1,*v2;


   if (argc < 2) {
      printf("Usage: %s ni\n",argv[0]);
      exit(0);
   } /* if */

   n=atoi(argv[1]);

   if (n > 1000000) {
      printf("\n");
printf("This program probably gives wrong result if n > 10^6 !!!\n");
      printf("\n");
      exit(0);
   } /* if */
   
   l1=n/4;
   l2=(int)(1.005*(double)(l1)+0.5)+10;
   n=4*l1;

   v1=(short *)calloc(l2+1,sizeof(short));
   v2=(short *)calloc(l2+1,sizeof(short));

   lt=1.0/log(10.0);

   l2f=4.0*(double)(l2);

   i=1;
   sum=0.0;

   while (sum < l2f) {
      sum=sum+lt*log((double)(i));
      i++;
   } /* while */

   fac=i;

   il=(int)(sum);

   printf("\n");
   printf("e = exp(1)\n");
   printf("==========\n");
   printf("\n");
   printf("Integer method, numerically stable.\n");
   printf("\n");
   printf("Factorial=%8d; Length=%8d; Decimals=%8d\n",fac,il,n);
   printf("\n");
   printf("Calculating ...\n");

   for (i=1; i<=l2; i++) {
      v1[i]=0;
      v2[i]=0;
   } /* for i */

   v1[l2]=1;
   v2[l2]=1;

   bas=10000;

   lld=0.0;

   time1=clock();

   for (j=1; j<=fac; j++) {

      jd=(double)(j);
      lld=lld+lt*log(jd);
      ll=(int)(0.25*lld)-6;

      if (ll > 0) {
         ll2=l2-ll;
      }
      else {
         ll2=l2;
      } /* if */


      rem=0;
      for (i=ll2; i>=1; i--) {
         q=(int)(v1[i])+bas*rem;
         v1[i]=(short)(q / j);
         rem=q % j;
      } /* for i */

      rem=0;
      for (i=1; i<=ll2; i++) {
         q=(int)(v1[i])+(int)(v2[i])+rem;
	 if (q >= bas) {
	    v2[i]=(short)(q-bas);
	    rem=1;
	 } 
	 else {
	    v2[i]=(short)(q);
	    rem=0;
	 } /* if */
      } /* for i */

      if (rem > 0) {
         for (i=ll2+1; i<=l2; i++) {
            q=(int)(v1[i])+(int)(v2[i])+rem;
	    if (q >= bas) {
	       v2[i]=(short)(q-bas);
	       rem=1;
	    } 
	    else {
	       v2[i]=(short)(q);
	       rem=0;
	    } /* if */
         } /* for i */
      } /* if */

   } /* for j */

   timediff1=clock()-time1;

   printf("\n");
   printf("CPU time=%10.2f s\n",(double)(timediff1)/timetic);
   printf("\n");
   printf("Calculation finished, output to %s.\n\n",fname);

   outfile=fopen(fname,"w");

   fprintf(outfile,"%d.\n",(int)(v2[l2]));

   imin=l2-l1;
   j=1;
   for (i=l2-1; i>=imin; i--) {
      fppp((int)(v2[i]),outfile);
      if ((j % 18) == 0) {
         fprintf(outfile,"\n");
      } /* if */
      j++;
   } /* for i */

   fclose(outfile);
   

} /* End */
